/*
 * ConvRBM_kernel.cu
 *
 *  Created on: Dec 25, 2009
 *      Author: Alex Krizhevsky (akrizhevsky@gmail.com)
 */
#include "ConvRBM_kernel.cuh"
#include "ConvRBM.cuh"

/*
 * Yeah this isn't very efficient but it's 1.4% of the cost of training so screw it.
 */
__global__ void kComputeSparseGrads(float* hidProbs, float* hidMeans, float* sparseGrads,
                                    const float sparseCost, const float sparseTarget,
                                    const int gradWidth, const int gradHeight, const int numConvOutputs) {

    const int ty = SPARSE_THREADS_Y * blockIdx.y + threadIdx.y;
    const int tx = SPARSE_THREADS_X * blockIdx.x + threadIdx.x;

    if (ty < gradHeight && tx < gradWidth) {
        const int gradIdx = ty * gradWidth + tx;
        const int meanIdx = ty * numConvOutputs + tx  % numConvOutputs;
        sparseGrads[gradIdx] = sparseCost * (hidMeans[meanIdx] - sparseTarget) * hidProbs[gradIdx] * (1 - hidProbs[gradIdx]);
    }
}

/*
 * hidProbs in column-major order (numHid x numCases)!
 */
__global__ void kComputeSparseGradsGlob(float* hidProbs, float* hidMeans, float sparseCost,
                                    float sparseTarget, float* sparseGrads,
                                    const int numCases, const int numHid) {
    const unsigned int idx = blockIdx.x * blockDim.x + threadIdx.x;

    for(unsigned int i = idx; i < numCases*numHid; i+= gridDim.x * blockDim.x) {
        sparseGrads[i] = sparseCost * (hidMeans[i / numCases] - sparseTarget) * hidProbs[i] * (1 - hidProbs[i]);
    }
}

/*
 * Block size (y,x) = (1,64)
 */
__global__ void kWeightNorm(float* hidVis, float* norms, const int height, const int width, const float thresh) {
    const int tx = WEIGHT_NORM_THREADS_X * blockIdx.x + threadIdx.x;
    const int ty = blockIdx.y ;
    __shared__ float norm;
    if (threadIdx.x == 0) {
        norm = norms[ty];
    }
    __syncthreads();
    if (tx < width && norm > thresh) {
        hidVis[ty * width + tx] *= thresh / norm;
    }
}

__global__ void kZeroNegatives(float* g, const int numElements) {
    const unsigned int idx = blockIdx.x * blockDim.x + threadIdx.x;

    for(unsigned int i = idx; i < numElements; i+= gridDim.x * blockDim.x) {
        if(g[i] < 0) {
            g[i] = 0;
        }
    }
}

__global__ void kComputeBiasGrads(float* hidProbs, float* derivs, const int sumLength, const int height, const int width) {
    const int y = (blockIdx.y * blockDim.y + threadIdx.y) * sumLength;
    const int x = blockIdx.x * blockDim.x + threadIdx.x;

    if(y < height && x < width) {
        float sum = 0;
        for(int i = y; i < y + sumLength; i++) {
            sum += hidProbs[i * width + x];
        }
        derivs[(y / sumLength) * width + x] = sum;
    }
}
